from __future__ import unicode_literals

import collections
import sys

from spotifyconnect import ffi, lib, serialized


PY2 = sys.version_info[0] == 2

if PY2:  # pragma: no branch
    string_types = (basestring,)  # noqa
    text_type = unicode  # noqa
    binary_type = str
else:
    string_types = (str,)
    text_type = str
    binary_type = bytes

class EventEmitter(object):

    """Mixin for adding event emitter functionality to a class."""

    def __init__(self):
        self._listeners = collections.defaultdict(list)

    @serialized
    def on(self, event, listener, *user_args):
        """Register a ``listener`` to be called on ``event``.

        The listener will be called with any extra arguments passed to
        :meth:`emit` first, and then the extra arguments passed to :meth:`on`
        last.

        If the listener function returns :class:`False`, it is removed and will
        not be called the next time the ``event`` is emitted.
        """
        self._listeners[event].append(
            _Listener(callback=listener, user_args=user_args))

    @serialized
    def off(self, event=None, listener=None):
        """Remove a ``listener`` that was to be called on ``event``.

        If ``listener`` is :class:`None`, all listeners for the given ``event``
        will be removed.

        If ``event`` is :class:`None`, all listeners for all events on this
        object will be removed.
        """
        if event is None:
            events = self._listeners.keys()
        else:
            events = [event]
        for event in events:
            if listener is None:
                self._listeners[event] = []
            else:
                self._listeners[event] = [
                    l for l in self._listeners[event]
                    if l.callback != listener]

    def emit(self, event, *event_args):
        """Call the registered listeners for ``event``.

        The listeners will be called with any extra arguments passed to
        :meth:`emit` first, and then the extra arguments passed to :meth:`on`
        """
        listeners = self._listeners[event][:]
        for listener in listeners:
            args = list(event_args) + list(listener.user_args)
            result = listener.callback(*args)
            if result is False:
                self.off(event, listener.callback)

    def num_listeners(self, event=None):
        """Return the number of listeners for ``event``.

        Return the total number of listeners for all events on this object if
        ``event`` is :class:`None`.
        """
        if event is not None:
            return len(self._listeners[event])
        else:
            return sum(len(l) for l in self._listeners.values())

    def call(self, event, *event_args):
        """Call the single registered listener for ``event``.

        The listener will be called with any extra arguments passed to
        :meth:`call` first, and then the extra arguments passed to :meth:`on`

        Raises :exc:`AssertionError` if there is none or multiple listeners for
        ``event``. Returns the listener's return value on success.
        """
        # XXX It would be a lot better for debugging if this error was raised
        # when registering the second listener instead of when the event is
        # emitted.
        assert self.num_listeners(event) == 1, (
            'Expected exactly 1 event listener, found %d listeners' %
            self.num_listeners(event))
        listener = self._listeners[event][0]
        args = list(event_args) + list(listener.user_args)
        return listener.callback(*args)


class _Listener(collections.namedtuple(
        'Listener', ['callback', 'user_args'])):

    """An listener of events from an :class:`EventEmitter`"""


class IntEnum(int):

    """An enum type for values mapping to integers.

    Tries to stay as close as possible to the enum type specified in
    :pep:`435` and introduced in Python 3.4.
    """

    def __new__(cls, value):
        if not hasattr(cls, '_values'):
            cls._values = {}
        if value not in cls._values:
            cls._values[value] = int.__new__(cls, value)
        return cls._values[value]

    def __repr__(self):
        if hasattr(self, '_name'):
            return '<%s.%s: %d>' % (self.__class__.__name__, self._name, self)
        else:
            return '<Unknown %s: %d>' % (self.__class__.__name__, self)

    @classmethod
    def add(cls, name, value):
        """Add a name-value pair to the enumeration."""
        attr = cls(value)
        attr._name = name
        setattr(cls, name, attr)


def make_enum(lib_prefix, enum_prefix=''):
    """Class decorator for automatically adding enum values.

    The values are read directly from the :attr:`spotify.lib` CFFI wrapper
    around libspotify. All values starting with ``lib_prefix`` are added. The
    ``lib_prefix`` is stripped from the name. Optionally, ``enum_prefix`` can
    be specified to add a prefix to all the names.
    """

    def wrapper(cls):
        for attr in dir(lib):
            if attr.startswith(lib_prefix):
                name = attr.replace(lib_prefix, enum_prefix)
                cls.add(name, getattr(lib, attr))
        return cls
    return wrapper


def to_bytes(value):
    """Converts bytes, unicode, and C char arrays to bytes.

    Unicode strings are encoded to UTF-8.
    """
    if isinstance(value, text_type):
        return value.encode('utf-8')
    elif isinstance(value, ffi.CData):
        return ffi.string(value)
    elif isinstance(value, binary_type):
        return value
    else:
        raise ValueError('Value must be text, bytes, or char[]')


def to_bytes_or_none(value):
    """Converts C char arrays to bytes and C NULL values to None."""
    if value == ffi.NULL:
        return None
    elif isinstance(value, ffi.CData):
        return ffi.string(value)
    else:
        raise ValueError('Value must be char[] or NULL')


def to_char(value):
    """Converts bytes, unicode, and C char arrays to C char arrays.  """
    return ffi.new('char[]', to_bytes(value))


def to_char_or_null(value):
    """Converts bytes, unicode, and C char arrays to C char arrays, and
    :class:`None` to C NULL values.
    """
    if value is None:
        return ffi.NULL
    else:
        return to_char(value)


def to_unicode(value):
    """Converts bytes, unicode, and C char arrays to unicode strings.

    Bytes and C char arrays are decoded from UTF-8.
    """
    if isinstance(value, ffi.CData):
        return ffi.string(value).decode('utf-8')
    elif isinstance(value, binary_type):
        return value.decode('utf-8')
    elif isinstance(value, text_type):
        return value
    else:
        raise ValueError('Value must be text, bytes, or char[]')


def to_unicode_or_none(value):
    """Converts C char arrays to unicode and C NULL values to None.

    C char arrays are decoded from UTF-8.
    """
    if value == ffi.NULL:
        return None
    elif isinstance(value, ffi.CData):
        return ffi.string(value).decode('utf-8')
    else:
        raise ValueError('Value must be char[] or NULL')



from __future__ import unicode_literals, division

import weakref

from spotifyconnect import ffi, lib, utils, serialized
import spotifyconnect


__all__ = [
    'Player',
    'PlayerEvent',
    'PlaybackNotify'
]

class Player(utils.EventEmitter):

    """Playback controller.

    You'll never need to create an instance of this class yourself. You'll find
    it ready to use as the :attr:`~Session.player` attribute on the
    :class:`Session` instance.
    """
    @serialized
    def __init__(self, session):
        super(Player, self).__init__()

        spotifyconnect._player_instance = self

        self._cache = weakref.WeakValueDictionary()
        self._emitters = []
        self._callback_handles = set()

        spotifyconnect.Error.maybe_raise(lib.SpRegisterPlaybackCallbacks(_PlayerCallbacks.get_struct(), session))

    _cache = None
    """A mapping from sp_* objects to their corresponding Python instances.

    The ``_cached`` helper constructors on wrapper objects use this cache for
    finding and returning existing alive wrapper objects for the sp_* object it
    is about to create a wrapper for.

    The cache *does not* keep objects alive. It's only a means for looking up
    the objects if they are kept alive somewhere else in the application.

    Internal attribute.
    """

    _emitters = None
    """A list of event emitters with attached listeners.

    When an event emitter has attached event listeners, we must keep the
    emitter alive for as long as the listeners are attached. This is achieved
    by adding them to this list.

    When creating wrapper objects around sp_* objects we must also return the
    existing wrapper objects instead of creating new ones so that the set of
    event listeners on the wrapper object can be modified. This is achieved
    with a combination of this list and the :attr:`_cache` mapping.

    Internal attribute.
    """

    _callback_handles = None
    """A set of handles returned by :meth:`spotify.ffi.new_handle`.

    These must be kept alive for the handle to remain valid until the callback
    arrives, even if the end user does not maintain a reference to the object
    the callback works on.

    Internal attribute.
    """

    @serialized
    def play(self):
        """Play the currently loaded track.

        This will cause alsa_sink data to be passed to the
        :attr:`~SessionCallbacks.music_delivery` callback.
        """
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackPlay())

    @serialized
    def pause(self):
        """Pause the currently loaded track.
        """
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackPause())

    @serialized
    def skip_to_next(self):
        """Skips to the next track on the playlist.
        """
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackSkipToNext())

    @serialized
    def skip_to_prev(self):
        """Skips to the previous track on the playlist.
        """
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackSkipToPrev())

    @serialized
    def seek(self, offset):
        """Seek to the offset in ms in the currently loaded track."""
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackSeek(offset))

    @serialized
    def enable_shuffle(self, value=None):
        """Enable shuffle mode
        """
        if value == None:
            value = not self.shuffled
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackEnableShuffle(value))

    @serialized
    def enable_repeat(self, value=None):
        """Enable repeat mode
        """
        if value == None:
            value = not self.repeated
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackEnableRepeat(value))

    @property
    @serialized
    def playing(self):
        return lib.SpPlaybackIsPlaying()

    @property
    @serialized
    def shuffled(self):
        return lib.SpPlaybackIsShuffled()

    @property
    @serialized
    def repeated(self):
        return lib.SpPlaybackIsRepeated()

    @property
    @serialized
    def active_device(self):
        return lib.SpPlaybackIsActiveDevice()

    @property
    @serialized
    def volume(self):
        return lib.SpPlaybackGetVolume()

    @volume.setter
    @serialized
    def volume(self, value):
        corrected_value = int(value * 655.35)
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackUpdateVolume(corrected_value))

    @property
    def current_track(self):
        return self.get_track_metadata()

    @property
    @serialized
    def metadata_valid_range(self):
        start = ffi.new("int *")
        end = ffi.new("int *")
        spotifyconnect.Error.maybe_raise(lib.SpGetMetadataValidRange(start, end))
        rango = {
            'start': start[0],
            'end': end[0]
        }
        return rango

    @serialized
    def get_track_metadata(self, offset=0):
        sp_metadata = ffi.new('SpMetadata *')
        spotifyconnect.Error.maybe_raise(lib.SpGetMetadata(sp_metadata, offset))
        return spotifyconnect.Metadata(sp_metadata)


    @serialized
    def set_bitrate(self, bitrate):
        spotifyconnect.Error.maybe_raise(lib.SpPlaybackSetBitrate(bitrate))

class PlayerEvent(object):

    """AlsaSink events.
    """
    PLAYBACK_NOTIFY = 'playback_notify'
    MUSIC_DELIVERY = 'playback_data'
    PLAYBACK_SEEK = 'playback_seek'
    PLAYBACK_VOLUME = 'playback_volume'

class _PlayerCallbacks(object):

    """Internal class."""

    @classmethod
    def get_struct(cls):
        return ffi.new('SpPlaybackCallbacks *', {
            'notify': cls.playback_notify,
            'audio_data': cls.playback_data,
            'seek': cls.playback_seek,
            'apply_volume':cls.playback_volume
        })

    # XXX Avoid use of the spotify._session_instance global in the following
    # callbacks.

    @staticmethod
    @ffi.callback('void(SpPlaybackNotify notify, void *userdata)')
    def playback_notify(sp_playback_notify, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        playback_notify = PlaybackNotify(sp_playback_notify)
        spotifyconnect._session_instance.player.emit(PlayerEvent.PLAYBACK_NOTIFY, playback_notify, ffi.from_handle(sp_userdata))

    @staticmethod
    @ffi.callback('uint32_t(void *data, uint32_t num_samples, SpSampleFormat *format, uint32_t *pending, void *userdata)')
    def playback_data(frames, num_frames, sp_audioformat, sp_pending, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        if spotifyconnect._session_instance.player.num_listeners(PlayerEvent.MUSIC_DELIVERY) == 0:
            return 0

        audio_format = spotifyconnect.AudioFormat(sp_audioformat)

        # Make sure waudio_formate don't pass incomplete frames
        num_frames -= num_frames % audio_format.channels

        frames_buffer = ffi.buffer(frames, num_frames * audio_format.frame_size)
        frames_bytes = frames_buffer[:]
        num_frames_consumed = spotifyconnect._session_instance.player.call(PlayerEvent.MUSIC_DELIVERY, audio_format, frames_bytes, num_frames, sp_pending, ffi.from_handle(sp_userdata))
        return num_frames_consumed

    @staticmethod
    @ffi.callback('void(uint32_t millis, void *userdata)')
    def playback_seek(sp_millis, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        millis = int(sp_millis)
        spotifyconnect._session_instance.player.emit(PlayerEvent.PLAYBACK_SEEK, millis, ffi.from_handle(sp_userdata))


    @staticmethod
    @ffi.callback('void(uint16_t volume, void *userdata)')
    def playback_volume(sp_volume, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        volume = sp_volume / 655.35
        spotifyconnect._session_instance.player.emit(PlayerEvent.PLAYBACK_VOLUME, volume, ffi.from_handle(sp_userdata))

@utils.make_enum('kSpPlaybackEvent')
@utils.make_enum('kSpPlaybackNotify')
class PlaybackNotify(utils.IntEnum):
    pass




from __future__ import unicode_literals

import spotifyconnect


class Sink(object):

    def on(self):
        """Turn on the alsa_sink sink.

        This is done automatically when the sink is instantiated, so you'll
        only need to call this method if you ever call :meth:`off` and want to
        turn the sink back on.
        """
        assert spotifyconnect._session_instance.player.num_listeners(
            spotifyconnect.PlayerEvent.MUSIC_DELIVERY) == 0
        spotifyconnect._session_instance.player.on(
            spotifyconnect.PlayerEvent.MUSIC_DELIVERY, self._on_music_delivery)

    def off(self):
        """Turn off the alsa_sink sink.

        This disconnects the sink from the relevant session events.
        """
        spotifyconnect._session_instance.player.off(
            spotifyconnect.PlayerEvent.MUSIC_DELIVERY, self._on_music_delivery)
        assert spotifyconnect._session_instance.player.num_listeners(
            spotifyconnect.PlayerEvent.MUSIC_DELIVERY) == 0
        self._close()

    def _on_music_delivery(self, audio_format, frames, num_frames, pending, session):
        # This method is called from an internal libspotify thread and must
        # not block in any way.
        raise NotImplementedError

    def _close(self):
        pass



from __future__ import unicode_literals

from multiprocessing import Process

from _zeroconfserver import app
from spotifyconnect import utils


__all__ = [
    'Zeroconf',
    'AvahiZeroConfServer'
]


class Zeroconf(object):

    """A Spotify ZeroConf object.
    """

    def __init__(self, sp_zeroconf):

        self.public_key = utils.to_unicode(sp_zeroconf.publicKey)
        self.device_id = utils.to_unicode(sp_zeroconf.deviceId)
        self.active_user = utils.to_unicode(sp_zeroconf.activeUser)
        self.remote_name = utils.to_unicode(sp_zeroconf.remoteName)
        self.account_req = utils.to_unicode(sp_zeroconf.accountReq)
        self.device_type = utils.to_unicode(sp_zeroconf.deviceType)


class AvahiZeroConfServer(Process):
    """A SpotifyConnect ZeroConf server.
    """

    daemon = True
    name = 'SpotifyConnectServer'

    def __init__(self, port=6697):
        Process.__init__(self)

        self._application = app
        self.port = port

    _application = None
    port = None

    def run(self):

        # First run the command avahi-publish-service SpotifyConnect _spotify-connect._tcp 6697 VERSION=1.0 CPath=/login/_zeroconf
        self._application.run(host='0.0.0.0', port=self.port, debug=True, use_reloader=False, threaded=True)



from __future__ import unicode_literals

import logging

from spotifyconnect import ffi, lib, utils
import spotifyconnect
from spotifyconnect.connection import error_callback


__all__ = [
    'Session'
]

logger = logging.getLogger(__name__)

class Session:

    """The Spotify session.

    If no ``config`` is provided, the default config is used.

    The session object will emit a number of events. See :class:`SessionEvent`
    for a list of all available events and how to connect your own listener
    functions up to get called when the events happens.

    .. warning::

        You can only have one :class:`Session` instance per process. This is a
        libspotify limitation. If you create a second :class:`Session` instance
        in the same process pyspotify-connect will raise a :exc:`RuntimeError` with the
        message "Session has already been initialized".

    :param config: the session config
    :type config: :class:`Config` or :class:`None`
    """

    def __init__(self, config=None):

        if spotifyconnect._session_instance is not None:
            raise RuntimeError('Session has already been initialized')

        userdata = ffi.new_handle(self)
        self._userdata = userdata

        if config is not None:
            self.config = config
        else:
            self.config = spotifyconnect.Config()

        self.config.error_callback = error_callback
        self.config.userdata = userdata

        spotifyconnect.Error.maybe_raise(lib.SpInit(self.config.sp_session_config))

        self.connection = spotifyconnect.Connection(userdata)
        self.player = spotifyconnect.Player(userdata)

        spotifyconnect._session_instance = self

    config = None
    """A :class:`Config` instance with the current configuration.

    Once the session has been created, changing the attributes of this object
    will generally have no effect.
    """

    connection = None
    """An :class:`~spotifyconnect.connection.Connection` instance for controlling the
    connection to the Spotify servers."""

    player = None
    """A :class:`~spotifyconnect.connection.Connection` instance for controlling playback."""

    _userdata = None
    """A internal variable to store the CData object handler for the :class:`~spotifyconnect.session.Session`"""

    def set_remote_name(self, remote_name):
        name = utils.to_char_or_null(remote_name or None)
        spotifyconnect.Error.maybe_raise(lib.SpSetDisplayName(name))

    def get_zeroconf_vars(self):
        zeroconf_vars = ffi.new('SpZeroConfVars *')
        spotifyconnect.Error.maybe_raise(lib.SpZeroConfGetVars(zeroconf_vars))
        zeroconf = spotifyconnect.Zeroconf(zeroconf_vars)
        return zeroconf

    def process_events(self):
        """Process pending events in libspotify.
        This method must be called for most callbacks to be called. Without
        calling this method, you'll only get the callbacks that are called from
        internal libspotify threads. When the
        pyspotify provides an :class:`~spotifyconnect.EventLoop` that you can use for
        processing events when needed.
        """
        spotifyconnect.Error.maybe_raise(lib.SpPumpEvents())

    @property
    def library_version(self):
        version = lib.SpGetLibraryVersion()
        return utils.to_unicode(version)

    def free_session(self):
        lib.SpFree()
        spotifyconnect._session_instance = None



import os

from distutils.version import StrictVersion

import cffi


if StrictVersion(cffi.__version__) < StrictVersion('1.0.0'):
    raise RuntimeError(
        'pyspotifyconnect requires cffi >= 1.0, but found %s' % cffi.__version__)


header_file = os.path.join(os.path.dirname(__file__), 'spotify.processed.h')

with open(header_file) as fh:
    header = fh.read()

ffi = cffi.FFI()
ffi.cdef(header)
ffi.cdef("""
void *malloc(size_t size);
void exit(int status);
""")

ffi.set_source(
    'spotifyconnect._spotifyconnect', '#include "spotify.h"', libraries=['spotify_embedded_shared'], include_dirs=[os.path.dirname(__file__)])

if __name__ == '__main__':
    ffi.compile()



from __future__ import unicode_literals

from spotifyconnect import ffi, lib, serialized, utils
import spotifyconnect


__all__ = [
    'ImageSize',
    'Metadata'
]


class Metadata(object):

    """A Spotify track.
    """

    def __init__(self, sp_metadata):

        self.playlist_name = utils.to_unicode(sp_metadata.data0)
        self.track_name = utils.to_unicode(sp_metadata.track_name)
        self.track_uri = utils.to_unicode(sp_metadata.track_uri)
        self.artist_name = utils.to_unicode(sp_metadata.artist_name)
        self.artist_uri = utils.to_unicode(sp_metadata.artist_uri)
        self.album_name = utils.to_unicode(sp_metadata.album_name)
        self.album_uri = utils.to_unicode(sp_metadata.album_uri)
        self.cover_uri = utils.to_unicode(sp_metadata.cover_uri)
        self.duration = sp_metadata.duration


    def __repr__(self):
        return 'Metadata(%r)' % self.track_uri

    @serialized
    def get_image_url(self, image_size):
        image_url = ffi.new('char[512]')
        spotifyconnect.Error.maybe_raise(lib.SpGetMetadataImageURL(utils.to_char(self.cover_uri), image_size, image_url, ffi.sizeof(image_url)))
        return utils.to_unicode(image_url)

@utils.make_enum('kSpImageSize')
class ImageSize(utils.IntEnum):
    pass



from __future__ import unicode_literals

import weakref

from spotifyconnect import ffi, lib, serialized, utils
import spotifyconnect


__all__ = [
    'Connection',
    'ConnectionEvent',
    'DebugEvent',
    'ConnectionState'
]

class Connection(utils.EventEmitter):

    """Connection controller.

    You'll never need to create an instance of this class yourself. You'll find
    it ready to use as the :attr:`~Session.connection` attribute on the
    :class:`Session` instance.
    """

    @serialized
    def __init__(self, session):
        super(Connection, self).__init__()

        self._connectionStatus = ConnectionState.LoggedOut
        spotifyconnect._connection_instance = self

        self._cache = weakref.WeakValueDictionary()
        self._emitters = []
        self._callback_handles = set()

        spotifyconnect.Error.maybe_raise(lib.SpRegisterConnectionCallbacks(_ConnectionCallbacks.get_struct(), session))
        spotifyconnect.Error.maybe_raise(lib.SpRegisterDebugCallbacks(_DebugCallbacks.get_struct(), session))

    @property
    @serialized
    def connection_state(self):
        """The session's current :class:`ConnectionState`.

        The mapping is as follows

        - :attr:`~ConnectionState.LoggedIn`: authenticated, online
        - :attr:`~ConnectionState.LoggedOut`: not authenticated
        - :attr:`~ConnectionState.TemporaryError`: Unknown error

        Register listeners for the
        :attr:`spotify.SessionEvent.CONNECTION_STATE_UPDATED` event to be
        notified when the connection state changes.
        """
        return ConnectionState(lib.SpConnectionIsLoggedIn())

    _cache = None
    """A mapping from sp_* objects to their corresponding Python instances.

    The ``_cached`` helper constructors on wrapper objects use this cache for
    finding and returning existing alive wrapper objects for the sp_* object it
    is about to create a wrapper for.

    The cache *does not* keep objects alive. It's only a means for looking up
    the objects if they are kept alive somewhere else in the application.

    Internal attribute.
    """

    _emitters = None
    """A list of event emitters with attached listeners.

    When an event emitter has attached event listeners, we must keep the
    emitter alive for as long as the listeners are attached. This is achieved
    by adding them to this list.

    When creating wrapper objects around sp_* objects we must also return the
    existing wrapper objects instead of creating new ones so that the set of
    event listeners on the wrapper object can be modified. This is achieved
    with a combination of this list and the :attr:`_cache` mapping.

    Internal attribute.
    """

    _callback_handles = None
    """A set of handles returned by :meth:`spotify.ffi.new_handle`.

    These must be kept alive for the handle to remain valid until the callback
    arrives, even if the end user does not maintain a reference to the object
    the callback works on.

    Internal attribute.
    """
    @serialized
    def login(self, username, password=None, blob=None, zeroconf=None):
        """Authenticate to Spotify's servers.

        You can login with one of three combinations:

        - ``username`` and ``password``
        - ``username`` and ``blob``
        - ``username`` and ``zeroconf``

        To get the ``blob`` string, you must once log in with ``username`` and
        ``password``. You'll then get the ``blob`` string passed to the
        :attr:`~ConnectionCallbacks.new_credentials` callback.
        """

        username = utils.to_char(username)

        if password is not None:
            spotifyconnect.Error.maybe_raise(lib.SpConnectionLoginPassword(username, password))
        elif blob is not None:
            spotifyconnect.Error.maybe_raise(lib.SpConnectionLoginBlob(username, blob))
        elif zeroconf is not None:
            spotifyconnect.Error.maybe_raise(lib.SpConnectionLoginZeroConf(username, *zeroconf))
        else:
            raise AttributeError("Must specify a login method (password, blob or zeroconf)")

    @serialized
    def logout(self):
        """Log out the current user.
        """
        spotifyconnect.Error.maybe_raise(lib.SpConnectionLogout())

class ConnectionEvent(object):

    """Connection events.
    """

    CONNECTION_NOTIFY_UPDATED = 'connection_notify'
    NEW_CREDENTIALS = 'connection_new_credentials'
    ERROR_NOTIFICATION = 'error_notification'

class DebugEvent(object):

    """Connection events.
    """

    DEBUG_MESSAGE = 'debug_message'


class _ConnectionCallbacks(object):

    """Internal class."""

    @classmethod
    def get_struct(cls):
        return ffi.new('SpConnectionCallbacks *', {
            'notify': cls.connection_notify,
            'new_credentials': cls.connection_new_credentials
        })

    # XXX Avoid use of the spotify._session_instance global in the following
    # callbacks.

    @staticmethod
    @ffi.callback('void(SpConnectionNotify type, void *userdata)')
    def connection_notify(sp_connection_notify, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        connection_notify = ConnectionState(sp_connection_notify)
        spotifyconnect._session_instance.connection.emit(
            ConnectionEvent.CONNECTION_NOTIFY_UPDATED,
            connection_notify, ffi.from_handle(sp_userdata))

    @staticmethod
    @ffi.callback('void(char *blob, void *userdata)')
    def connection_new_credentials(sp_blob, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        blob = utils.to_unicode(sp_blob)
        spotifyconnect._session_instance.connection.emit(
            ConnectionEvent.NEW_CREDENTIALS, blob, ffi.from_handle(sp_userdata))

class _DebugCallbacks(object):

    """Internal class."""

    @classmethod
    def get_struct(cls):
        return ffi.new('SpDebugCallbacks *', {
            'message': cls.debug_message
        })

    # XXX Avoid use of the spotify._session_instance global in the following
    # callbacks.

    @staticmethod
    @ffi.callback('void(char *msg, void *userdata)')
    def debug_message(sp_message, sp_userdata):
        if not spotifyconnect._session_instance:
            return
        message = utils.to_unicode(sp_message)
        spotifyconnect._session_instance.connection.emit(
            DebugEvent.DEBUG_MESSAGE,
            message, ffi.from_handle(sp_userdata))

@utils.make_enum('kSpConnectionNotify')
class ConnectionState(utils.IntEnum):
    pass

# Error callbacks
@ffi.callback('void(SpError error, void *userdata)')
def error_callback(error, sp_userdata):
    spotifyconnect._session_instance.connection.emit(ConnectionEvent.ERROR_NOTIFICATION,
        spotifyconnect.ErrorType(error), ffi.from_handle(sp_userdata))




from __future__ import unicode_literals

import uuid

from spotifyconnect import ffi, utils


__all__ = [
    'Config',
    'DeviceType'
]


class Config(object):

    """The session config.

    Create an instance and assign to its attributes to configure. Then use the
    config object to create a session::

        >>> config = spotify.Config()
        >>> config.user_agent = 'My awesome Spotify client'
        >>> # Etc ...
        >>> session = spotify.Session(config=config)
    """

    def __init__(self):
        self._sp_session_config = ffi.new('SpConfig *')

        self.version = 4
        self._sp_session_config.buffer = ffi.dlopen(None).malloc(0x100000)
        self._sp_session_config.buffer_size = 0x100000
        self.load_application_key_file()
        self.device_id = str(uuid.uuid4())
        self.remote_name = 'Spotify-Connect'
        self.brand_name = 'DummyBrand'
        self.model_name = 'DummyModel'
        self.device_type = DeviceType.AudioDongle
        self.userdata = ffi.new_handle(self)

    @property
    def sp_session_config(self):
        return self._sp_session_config

    @property
    def version(self):
        """The API version of the libspotify we're using.

        You should not need to change this. It defaults to the value provided
        by libspotify through :func:`spotify.get_libspotify_api_version`.
        """
        return self._sp_session_config.version

    @version.setter
    def version(self, value):
        self._sp_session_config.version = value

    @property
    def app_key(self):
        """Your libspotify application key.

        Must be a bytestring. Alternatively, you can call
        :meth:`load_application_key_file`, and pyspotify will correctly read
        the file into :attr:`application_key`.
        """
        return utils.to_bytes_or_none(
            ffi.cast('char *', self._sp_session_config.app_key))

    @app_key.setter
    def app_key(self, value):
        if value is None:
            size = 0
        else:
            size = len(value)
        assert size in (0, 321), (
            'Invalid application key; expected 321 bytes, got %d bytes' % size)

        self._application_key = utils.to_char_or_null(value)
        self._sp_session_config.app_key = ffi.cast(
            'void *', self._application_key)
        self._sp_session_config.app_key_size = size

    def load_application_key_file(self, filename=b'spotify_appkey.key'):
        """Load your libspotify application key file.

        If called without arguments, it tries to read ``spotify_appkey.key``
        from the current working directory.

        This is an alternative to setting :attr:`application_key` yourself. The
        file must be a binary key file, not the C code key file that can be
        compiled into an application.
        """
        with open(filename, 'rb') as fh:
            self.app_key = fh.read()

    @property
    def device_id(self):
        """Device ID for offline synchronization and logging purposes.

        Defaults to :class:`None`.

        The Device ID must be unique to the particular device instance, i.e. no
        two units must supply the same Device ID. The Device ID must not change
        between sessions or power cycles. Good examples is the device's MAC
        address or unique serial number.

        Setting the device ID to an empty string has the same effect as setting
        it to :class:`None`.
        """
        return utils.to_unicode_or_none(self._sp_session_config.deviceId)

    @device_id.setter
    def device_id(self, value):
        # NOTE libspotify segfaults if device_id is set to an empty string,
        # thus we convert empty strings to NULL.
        self._deviceId = utils.to_char_or_null(value or None)
        self._sp_session_config.deviceId = self._deviceId

    @property
    def remote_name(self):
        return utils.to_unicode_or_none(self._sp_session_config.remoteName)

    @remote_name.setter
    def remote_name(self, value):
        self._remoteName = utils.to_char_or_null(value or None)
        self._sp_session_config.remoteName = self._remoteName

    @property
    def brand_name(self):
        return utils.to_unicode_or_none(self._sp_session_config.brandName)

    @brand_name.setter
    def brand_name(self, value):
        self._brandName = utils.to_char_or_null(value or None)
        self._sp_session_config.brandName = self._brandName

    @property
    def model_name(self):
        return utils.to_unicode_or_none(self._sp_session_config.modelName)

    @model_name.setter
    def model_name(self, value):
        self._modelName = utils.to_char_or_null(value or None)
        self._sp_session_config.modelName = self._modelName

    @property
    def device_type(self):
        return DeviceType(self._sp_session_config.deviceType)

    @device_type.setter
    def device_type(self, value):
        self._sp_session_config.deviceType = value

    @property
    def userdata(self):
        return self._sp_session_config.userdata

    @userdata.setter
    def userdata(self, value):
        self._sp_session_config.userdata = value

    @property
    def error_callback(self):
        return self._sp_session_config.error_callback

    @error_callback.setter
    def error_callback(self, value):
        self._sp_session_config.error_callback = value

@utils.make_enum('kSpDeviceType')
class DeviceType(utils.IntEnum):
    pass



from __future__ import unicode_literals

import threading

__version__ = '0.1.2'


# Global reentrant lock to be held whenever libspotify functions are called or
# libspotify owned data is worked on. This is the heart of pyspotify's thread
# safety.
_lock = threading.RLock()


# Reference to the spotifyconnect.Session instance. Used to enforce that one and only
# one session exists in each process.
_session_instance = None

def _setup_logging():
    """Setup logging to log to nowhere by default.

    For details, see:
    http://docs.python.org/3/howto/logging.html#library-config

    Internal function.
    """
    import logging

    logger = logging.getLogger('spotify-connect')
    handler = logging.NullHandler()
    logger.addHandler(handler)


def serialized(f):
    """Decorator that serializes access to all decorated functions.

    The decorator acquires pyspotify's single global lock while calling any
    wrapped function. It is used to serialize access to:

    - All calls to functions on :attr:`spotify.lib`.

    - All code blocks working on pointers returned from functions on
      :attr:`spotify.lib`.

    - All code blocks working on other internal data structures in pyspotify.

    Together this is what makes pyspotify safe to use from multiple threads and
    enables convenient features like the :class:`~spotify.EventLoop`.

    Internal function.
    """
    import functools

    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        with _lock:
            return f(*args, **kwargs)
    if not hasattr(wrapper, '__wrapped__'):
        # Workaround for Python < 3.2
        wrapper.__wrapped__ = f
    return wrapper


class _SerializedLib(object):
    """CFFI library wrapper to serialize all calls to library functions.

    Internal class.
    """

    def __init__(self, lib):
        for name in dir(lib):
            attr = getattr(lib, name)
            if name.startswith('SP') and callable(attr):
                attr = serialized(attr)
            setattr(self, name, attr)


_setup_logging()

from spotifyconnect._spotifyconnect import ffi, lib  # noqa

lib = _SerializedLib(lib)

from spotifyconnect.audio import *  # noqa
from spotifyconnect.config import *  # noqa
from spotifyconnect.connection import *  # noqa
from spotifyconnect.error import *  # noqa
from spotifyconnect.eventloop import *  # noqa
from spotifyconnect.metadata import *  # noqa
from spotifyconnect.player import *  # noqa
from spotifyconnect.session import *  # noqa
from spotifyconnect.sink import *  # noqa
from spotifyconnect.zeroconf import *  # noqa



from flask import Flask, request, jsonify
import spotifyconnect


app = Flask('SpotifyConnect')

# #API routes
# Login routes
@app.route('/login/_zeroconf', methods=['GET', 'POST'])
def login_zeroconf():
    action = request.args.get('action') or request.form.get('action')
    if not action:
        return jsonify({
            'status': 301,
            'spotifyError': 0,
            'statusString': 'ERROR-MISSING-ACTION'})
    if action == 'getInfo' and request.method == 'GET':
        return get_info()
    elif action == 'addUser' and request.method == 'POST':
        return add_user()
    else:
        return jsonify({
            'status': 301,
            'spotifyError': 0,
            'statusString': 'ERROR-INVALID-ACTION'})

def get_info():
    zeroconf_vars = spotifyconnect._session_instance.get_zeroconf_vars()

    return jsonify({
        'status': 101,
        'spotifyError': 0,
        'activeUser': zeroconf_vars.active_user,
        'brandDisplayName': spotifyconnect._session_instance.config.brand_name,
        'accountReq': zeroconf_vars.account_req,
        'deviceID': zeroconf_vars.device_id,
        'publicKey': zeroconf_vars.public_key,
        'version': '2.0.1',
        'deviceType': zeroconf_vars.device_type,
        'modelDisplayName': spotifyconnect._session_instance.config.model_name,
        # Status codes are ERROR-OK (not actually an error), ERROR-MISSING-ACTION, ERROR-INVALID-ACTION, ERROR-SPOTIFY-ERROR, ERROR-INVALID-ARGUMENTS, ERROR-UNKNOWN, and ERROR_LOG_FILE
        'statusString': 'ERROR-OK',
        'remoteName': zeroconf_vars.remote_name,
    })

def add_user():
    args = request.form
    # TODO: Add parameter verification
    username = str(args.get('userName'))
    blob = str(args.get('blob'))
    clientKey = str(args.get('clientKey'))

    spotifyconnect._session_instance.connection.login(username, zeroconf=(blob, clientKey))

    return jsonify({
        'status': 101,
        'spotifyError': 0,
        'statusString': 'ERROR-OK'
        })



from __future__ import unicode_literals

import re


from setuptools import find_packages, setup


def read_file(filename):
    with open(filename) as fh:
        return fh.read()


def get_version(filename):
    init_py = read_file(filename)
    metadata = dict(re.findall("__([a-z]+)__ = '([^']+)'", init_py))
    return metadata['version']


setup(
    name='pyspotify-connect',
    version=get_version('spotifyconnect/__init__.py'),
    url='https://github.com/chukysoria/pyspotify-connect',
    license='Apache License, Version 2.0',
    author='chukysoria',
    author_email='nomail@nomail.com',
    description='Python wrapper for libspotify-connect',
    long_description=read_file('README.rst'),
    keywords='spotify connect library',
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    setup_requires=['cffi >= 1.0.0'],
    cffi_modules=['spotifyconnect/_spotifyconnect_build.py:ffi'],
    install_requires=[
        'cffi >= 1.0.0',
        'Flask >= 0.10.1'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
    ],
)
